package vazkii.quark.experimental.entity;

import com.google.common.collect.Sets;
import net.minecraft.entity.EntityAgeable;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.ai.*;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.passive.EntityAnimal;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.init.Items;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.datasync.DataParameter;
import net.minecraft.network.datasync.DataSerializers;
import net.minecraft.network.datasync.EntityDataManager;
import net.minecraft.pathfinding.Path;
import net.minecraft.util.*;
import net.minecraft.util.math.MathHelper;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;
import vazkii.quark.base.sounds.QuarkSounds;
import vazkii.quark.experimental.features.Frogs;
import vazkii.quark.world.entity.ai.EntityAIFavorBlock;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Calendar;
import java.util.Set;

public class EntityFrog extends EntityAnimal {

	public static final ResourceLocation FROG_LOOT_TABLE = new ResourceLocation("quark", "entities/frog");

	private static final DataParameter<Integer> TALK_TIME = EntityDataManager.createKey(EntityFrog.class, DataSerializers.VARINT);
	private static final Set<Item> TEMPTATION_ITEMS = Sets.newHashSet(Items.FISH, Items.SPIDER_EYE);

	public int spawnCd = -1;
	public int spawnChain = 30;

	public EntityFrog(World worldIn) {
		super(worldIn);
		setSize(0.9f, 0.5f);

		this.jumpHelper = new FrogJumpHelper();
		this.moveHelper = new FrogMoveHelper();
		this.setMovementSpeed(0.0D);
	}

	@Override
	protected void entityInit() {
		super.entityInit();

		dataManager.register(TALK_TIME, 0);
	}

	@Override
	protected void initEntityAI() {
		tasks.addTask(0, new EntityAISwimming(this));
		tasks.addTask(1, new AIPanic(1.25));
		tasks.addTask(2, new EntityAIMate(this, 1.0));
		tasks.addTask(3, new EntityAITempt(this, 1.2, false, TEMPTATION_ITEMS));
		tasks.addTask(4, new EntityAIFollowParent(this, 1.1));
		tasks.addTask(5, new EntityAIFavorBlock(this, 1, Blocks.WATERLILY));
		tasks.addTask(6, new EntityAIWanderAvoidWater(this, 1, 0.5F));
		tasks.addTask(7, new EntityAIWatchClosest(this, EntityPlayer.class, 6));
		tasks.addTask(8, new EntityAILookIdle(this));
	}

	@Override
	protected void applyEntityAttributes() {
		super.applyEntityAttributes();
		this.getEntityAttribute(SharedMonsterAttributes.MAX_HEALTH).setBaseValue(10);
		this.getEntityAttribute(SharedMonsterAttributes.MOVEMENT_SPEED).setBaseValue(0.25);
	}

	public int getTalkTime() {
		return dataManager.get(TALK_TIME);
	}

	@Override
	public void onUpdate() {
		super.onUpdate();

		int talkTime = getTalkTime();
		if (talkTime > 0)
			dataManager.set(TALK_TIME, talkTime - 1);

		if (Frogs.frogsDoTheFunny && spawnCd > 0 && spawnChain > 0) {
			spawnCd--;
			if (spawnCd == 0 && !world.isRemote) {
				float multiplier = 0.8F;
				EntityFrog newFrog = new EntityFrog(world);
				newFrog.setPosition(posX, posY, posZ);
				newFrog.motionX = (Math.random() - 0.5) * multiplier;
				newFrog.motionY = (Math.random() - 0.5) * multiplier;
				newFrog.motionZ = (Math.random() - 0.5) * multiplier;
				world.spawnEntity(newFrog);
				newFrog.spawnCd = 2;
				newFrog.spawnChain = spawnChain - 1;
				spawnChain = 0;
			}
		}

		this.prevRotationYaw = this.prevRotationYawHead;
		this.rotationYaw = this.rotationYawHead;
	}

	@Override
	protected boolean canDropLoot() {
		return spawnChain != 0;
	}

	@Nullable
	@Override
	protected ResourceLocation getLootTable() {
		return FROG_LOOT_TABLE;
	}

	private int droppedLegs = -1;

	@Override
	protected void dropLoot(boolean wasRecentlyHit, int lootingModifier, @Nonnull DamageSource source) {
		droppedLegs = 0;
		super.dropLoot(wasRecentlyHit, lootingModifier, source);
		droppedLegs = -1;
	}

	@Nullable
	@Override
	public EntityItem entityDropItem(ItemStack stack, float offsetY) {
		if (droppedLegs >= 0) {
			int count = Math.max(4 - droppedLegs, 0);
			droppedLegs += stack.getCount();

			if (stack.getCount() > count) {
				ItemStack copy = stack.copy();
				copy.shrink(count);
				copy.getOrCreateSubCompound("display")
						.setString("LocName", "item.quark:frog_maybe_leg.name");

				stack = stack.copy();
				stack.shrink(copy.getCount());

				super.entityDropItem(copy, offsetY);
			}
		}

		return super.entityDropItem(stack, offsetY);
	}

	@Override
	public boolean processInteract(EntityPlayer player, @Nonnull EnumHand hand) {
		if (super.processInteract(player, hand))
			return true;

		Calendar calendar = world.getCurrentDate();
		if (Frogs.frogsDoTheFunny && calendar.get(Calendar.DAY_OF_WEEK) == Calendar.WEDNESDAY) {
			if (!world.isRemote) if (spawnChain > 0) {
				spawnCd = 50;
				dataManager.set(TALK_TIME, 80);
				world.playSound(null, posX, posY, posZ, QuarkSounds.ENTITY_FROG_WEDNESDAY, SoundCategory.NEUTRAL, 1F, 1F);
			}

			return true;
		}

		return false;
	}

	@Nullable
	@Override
	public EntityAgeable createChild(@Nonnull EntityAgeable otherParent) {
		return new EntityFrog(world);
	}

	@Override
	public boolean isBreedingItem(ItemStack stack) {
		return !stack.isEmpty() && TEMPTATION_ITEMS.contains(stack.getItem());
	}

	@Override
	public void readEntityFromNBT(NBTTagCompound compound) {
		super.readEntityFromNBT(compound);
		spawnCd = compound.getInteger("Cooldown");
		spawnChain = compound.getInteger("Chain");
		dataManager.set(TALK_TIME, compound.getInteger("DudeAmount"));
	}

	@Override
	public void writeEntityToNBT(NBTTagCompound compound) {
		super.writeEntityToNBT(compound);
		compound.setInteger("Cooldown", spawnCd);
		compound.setInteger("Chain", spawnChain);
		compound.setInteger("DudeAmount", getTalkTime());
	}

	@Override
	protected SoundEvent getAmbientSound() {
		return QuarkSounds.ENTITY_FROG_IDLE;
	}

	@Override
	protected SoundEvent getHurtSound(DamageSource damageSourceIn) {
		return QuarkSounds.ENTITY_FROG_HURT;
	}

	@Override
	protected SoundEvent getDeathSound() {
		return QuarkSounds.ENTITY_FROG_DIE;
	}

	protected SoundEvent getJumpSound() {
		return QuarkSounds.ENTITY_FROG_JUMP;
	}

	// Begin copypasta from EntityRabbit

	private int jumpTicks;
	private int jumpDuration;
	private boolean wasOnGround;
	private int currentMoveTypeDuration;

	@Override
	public void updateAITasks() {
		if (this.currentMoveTypeDuration > 0) --this.currentMoveTypeDuration;

		if (this.onGround) {
			if (!this.wasOnGround) {
				this.setJumping(false);
				this.checkLandingDelay();
			}

			FrogJumpHelper jumpHelper = (FrogJumpHelper) this.jumpHelper;

			if (!jumpHelper.getIsJumping()) {
				if (this.moveHelper.isUpdating() && this.currentMoveTypeDuration == 0) {
					Path path = this.navigator.getPath();
					Vec3d vec3d = new Vec3d(this.moveHelper.getX(), this.moveHelper.getY(), this.moveHelper.getZ());

					if (path != null && path.getCurrentPathIndex() < path.getCurrentPathLength())
						vec3d = path.getPosition(this);

					this.calculateRotationYaw(vec3d.x, vec3d.z);
					this.startJumping();
				}
			} else if (!jumpHelper.canJump()) this.enableJumpControl();
		}

		this.wasOnGround = this.onGround;
	}

	@Override
	public void spawnRunningParticles() {
		// NO-OP
	}

	private void calculateRotationYaw(double x, double z) {
		this.rotationYaw = (float) (MathHelper.atan2(z - this.posZ, x - this.posX) * (180D / Math.PI)) - 90.0F;
	}

	private void enableJumpControl() {
		((FrogJumpHelper) this.jumpHelper).setCanJump(true);
	}

	private void disableJumpControl() {
		((FrogJumpHelper) this.jumpHelper).setCanJump(false);
	}

	private void updateMoveTypeDuration() {
		if (this.moveHelper.getSpeed() < 2.2D)
			this.currentMoveTypeDuration = 10;
		else
			this.currentMoveTypeDuration = 1;
	}

	private void checkLandingDelay() {
		this.updateMoveTypeDuration();
		this.disableJumpControl();
	}

	@Override
	public void onLivingUpdate() {
		super.onLivingUpdate();

		if (this.jumpTicks != this.jumpDuration) ++this.jumpTicks;
		else if (this.jumpDuration != 0) {
			this.jumpTicks = 0;
			this.jumpDuration = 0;
			this.setJumping(false);
		}
	}

	@Override
	protected void jump() {
		super.jump();
		double d0 = this.moveHelper.getSpeed();

		if (d0 > 0.0D) {
			double d1 = this.motionX * this.motionX + this.motionZ * this.motionZ;

			if (d1 < 0.01) this.moveRelative(0.0F, 0.0F, 1.0F, 0.1F);
		}

		if (!this.world.isRemote)
			this.world.setEntityState(this, (byte) 1);
	}

	public void setMovementSpeed(double newSpeed) {
		this.getNavigator().setSpeed(newSpeed);
		this.moveHelper.setMoveTo(this.moveHelper.getX(), this.moveHelper.getY(), this.moveHelper.getZ(), newSpeed);
	}

	@Override
	public void setJumping(boolean jumping) {
		super.setJumping(jumping);

		if (jumping)
			this.playSound(this.getJumpSound(), this.getSoundVolume(), ((this.rand.nextFloat() - this.rand.nextFloat()) * 0.2F + 1.0F) * 0.8F);
	}

	public void startJumping() {
		this.setJumping(true);
		this.jumpDuration = 10;
		this.jumpTicks = 0;
	}

	@Override
	@SideOnly(Side.CLIENT)
	public void handleStatusUpdate(byte id) {
		if (id == 1) {
			this.createRunningParticles();
			this.jumpDuration = 10;
			this.jumpTicks = 0;
		} else
			super.handleStatusUpdate(id);
	}

	public class FrogJumpHelper extends EntityJumpHelper {
		private boolean canJump;

		public FrogJumpHelper() {
			super(EntityFrog.this);
		}

		public boolean getIsJumping() {
			return this.isJumping;
		}

		public boolean canJump() {
			return this.canJump;
		}

		public void setCanJump(boolean canJumpIn) {
			this.canJump = canJumpIn;
		}

		@Override
		public void doJump() {
			if (this.isJumping) {
				startJumping();
				this.isJumping = false;
			}
		}
	}

	public class FrogMoveHelper extends EntityMoveHelper {
		private double nextJumpSpeed;

		public FrogMoveHelper() {
			super(EntityFrog.this);
		}

		@Override
		public void onUpdateMoveHelper() {
			if (onGround && !isJumping && !((EntityFrog.FrogJumpHelper) jumpHelper).getIsJumping())
				setMovementSpeed(0.0D);
			else if (this.isUpdating()) setMovementSpeed(this.nextJumpSpeed);

			super.onUpdateMoveHelper();
		}

		@Override
		public void setMoveTo(double x, double y, double z, double speedIn) {
			if (isInWater()) speedIn = 1.5D;

			super.setMoveTo(x, y, z, speedIn);

			if (speedIn > 0.0D) this.nextJumpSpeed = speedIn;
		}
	}

	public class AIPanic extends EntityAIPanic {

		public AIPanic(double speedIn) {
			super(EntityFrog.this, speedIn);
		}

		public void updateTask() {
			super.updateTask();
			setMovementSpeed(this.speed);
		}
	}
}
